@@ -70,6 +70,12 @@ int ExternalFilePtr_gets(
 	int *EOL_in_buf
 );
 
+void ExternalFilePtr_seek(
+	SEXP efp,
+	long long int offset,
+	int whence
+);
+
 void ExternalFilePtr_rewind(SEXP efp);
 
 int ExternalFilePtr_puts(
@@ -230,21 +236,19 @@ SEXP XStringSet_xscat(SEXP args);
 
 SEXP debug_XStringSet_io();
 
-SEXP fasta_info(
+SEXP fasta_index(
 	SEXP efp_list,
 	SEXP nrec,
 	SEXP skip,
 	SEXP seek_first_rec,
-	SEXP use_names,
 	SEXP lkup
 );
 
-SEXP read_fasta_in_XStringSet(
+SEXP read_XStringSet_from_fasta_blocks(
+	SEXP seqlength,
 	SEXP efp_list,
-	SEXP nrec,
-	SEXP skip,
-	SEXP seek_first_rec,
-	SEXP use_names,
+	SEXP nrec_list,
+	SEXP offset_list,
 	SEXP elementType,
 	SEXP lkup
 );
@@ -263,7 +267,7 @@ SEXP fastq_geometry(
 	SEXP seek_first_rec
 );
 
-SEXP read_fastq_in_XStringSet(
+SEXP read_XStringSet_from_fastq(
 	SEXP efp_list,
 	SEXP nrec,
 	SEXP skip,
@@ -847,11 +851,16 @@ SEXP XStringViews_match_PWM(
 SEXP debug_find_palindromes();
 
 SEXP find_palindromes(
-	SEXP s_xp,
-	SEXP s_offset,
-	SEXP s_length,
+	SEXP x,
 	SEXP min_armlength,
 	SEXP max_looplength,
+	SEXP max_mismatch,
+	SEXP L2R_lkup
+);
+
+SEXP palindrome_arm_length(
+	SEXP x,
+	SEXP max_mismatch,
 	SEXP L2R_lkup
 );
 
