@@ -56,11 +56,11 @@ static const R_CallMethodDef callMethods[] = {
 
 /* XStringSet_io.c */
 	CALLMETHOD_DEF(debug_XStringSet_io, 0),
-	CALLMETHOD_DEF(fasta_info, 6),
-	CALLMETHOD_DEF(read_fasta_in_XStringSet, 7),
+	CALLMETHOD_DEF(fasta_index, 5),
+	CALLMETHOD_DEF(read_XStringSet_from_fasta_blocks, 6),
 	CALLMETHOD_DEF(write_XStringSet_to_fasta, 4),
 	CALLMETHOD_DEF(fastq_geometry, 4),
-	CALLMETHOD_DEF(read_fastq_in_XStringSet, 7),
+	CALLMETHOD_DEF(read_XStringSet_from_fastq, 7),
 	CALLMETHOD_DEF(write_XStringSet_to_fastq, 4),
 
 /* letter_frequency.c */
@@ -140,7 +140,8 @@ static const R_CallMethodDef callMethods[] = {
 
 /* find_palindromes.c */
 	CALLMETHOD_DEF(debug_find_palindromes, 0),
-	CALLMETHOD_DEF(find_palindromes, 6),
+	CALLMETHOD_DEF(find_palindromes, 5),
+	CALLMETHOD_DEF(palindrome_arm_length, 3),
 
 /* BitMatrix.c */
 	CALLMETHOD_DEF(debug_BitMatrix, 0),
