@@ -3,8 +3,10 @@
  *                           Author: Herve Pages                            *
  ****************************************************************************/
 #include "Biostrings.h"
-#include "IRanges_interface.h"
 #include "XVector_interface.h"
+#include "S4Vectors_interface.h"
+
+#include <math.h>  /* for llround */
 
 static int debug = 0;
 
@@ -57,6 +59,7 @@ typedef struct fasta_loader {
 	const int *lkup;
 	int lkup_length;
 	void (*load_desc_line)(struct fasta_loader *loader,
+			       int recno, long long int offset,
 			       const Chars_holder *desc_line);
 	void (*load_empty_seq)(struct fasta_loader *loader);
 	void (*load_seq_data)(struct fasta_loader *loader,
@@ -66,63 +69,78 @@ typedef struct fasta_loader {
 } FASTAloader;
 
 /*
- * The FASTAINFO loader only loads the lengths (and optionally the names)
- * of the sequences.
+ * The FASTAINDEX loader.
  */
 
-typedef struct fastainfo_loader_ext {
-	CharAEAE ans_names_buf;
-	IntAE seqlengths_buf;
-} FASTAINFO_loaderExt;
+typedef struct fastaindex_loader_ext {
+	IntAE recno_buf;
+	LongLongIntAE offset_buf;
+	CharAEAE desc_buf;
+	IntAE seqlength_buf;
+} FASTAINDEX_loaderExt;
 
-static void FASTAINFO_load_desc_line(FASTAloader *loader,
-		const Chars_holder *desc_line)
+static FASTAINDEX_loaderExt new_FASTAINDEX_loaderExt()
 {
-	FASTAINFO_loaderExt *loader_ext;
-	CharAEAE *ans_names_buf;
+	FASTAINDEX_loaderExt loader_ext;
+
+	loader_ext.recno_buf = new_IntAE(0, 0, 0);
+	loader_ext.offset_buf = new_LongLongIntAE(0, 0, 0);
+	loader_ext.desc_buf = new_CharAEAE(0, 0);
+	loader_ext.seqlength_buf = new_IntAE(0, 0, 0);
+	return loader_ext;
+}
+
+static void FASTAINDEX_load_desc_line(FASTAloader *loader,
+				      int recno, long long int offset,
+				      const Chars_holder *desc_line)
+{
+	FASTAINDEX_loaderExt *loader_ext;
+	IntAE *recno_buf;
+	LongLongIntAE *offset_buf;
+	CharAEAE *desc_buf;
 
 	loader_ext = loader->ext;
-	ans_names_buf = &(loader_ext->ans_names_buf);
+
+	recno_buf = &(loader_ext->recno_buf);
+	IntAE_insert_at(recno_buf, IntAE_get_nelt(recno_buf), recno + 1);
+
+	offset_buf = &(loader_ext->offset_buf);
+	LongLongIntAE_insert_at(offset_buf,
+				LongLongIntAE_get_nelt(offset_buf),
+				offset);
+
+	desc_buf = &(loader_ext->desc_buf);
 	// This works only because desc_line->seq is nul-terminated!
-	append_string_to_CharAEAE(ans_names_buf, desc_line->seq);
+	append_string_to_CharAEAE(desc_buf, desc_line->ptr);
 	return;
 }
 
-static void FASTAINFO_load_empty_seq(FASTAloader *loader)
+static void FASTAINDEX_load_empty_seq(FASTAloader *loader)
 {
-	FASTAINFO_loaderExt *loader_ext;
-	IntAE *seqlengths_buf;
+	FASTAINDEX_loaderExt *loader_ext;
+	IntAE *seqlength_buf;
 
 	loader_ext = loader->ext;
-	seqlengths_buf = &(loader_ext->seqlengths_buf);
-	IntAE_insert_at(seqlengths_buf, IntAE_get_nelt(seqlengths_buf), 0);
+	seqlength_buf = &(loader_ext->seqlength_buf);
+	IntAE_insert_at(seqlength_buf, IntAE_get_nelt(seqlength_buf), 0);
 	return;
 }
 
-static void FASTAINFO_load_seq_data(FASTAloader *loader,
+static void FASTAINDEX_load_seq_data(FASTAloader *loader,
 		const Chars_holder *seq_data)
 {
-	FASTAINFO_loaderExt *loader_ext;
-	IntAE *seqlengths_buf;
+	FASTAINDEX_loaderExt *loader_ext;
+	IntAE *seqlength_buf;
 
 	loader_ext = loader->ext;
-	seqlengths_buf = &(loader_ext->seqlengths_buf);
-	seqlengths_buf->elts[IntAE_get_nelt(seqlengths_buf) - 1] +=
+	seqlength_buf = &(loader_ext->seqlength_buf);
+	seqlength_buf->elts[IntAE_get_nelt(seqlength_buf) - 1] +=
 		seq_data->length;
 	return;
 }
 
-static FASTAINFO_loaderExt new_FASTAINFO_loaderExt()
-{
-	FASTAINFO_loaderExt loader_ext;
-
-	loader_ext.ans_names_buf = new_CharAEAE(0, 0);
-	loader_ext.seqlengths_buf = new_IntAE(0, 0, 0);
-	return loader_ext;
-}
-
-static FASTAloader new_FASTAINFO_loader(SEXP lkup, int load_descs,
-		FASTAINFO_loaderExt *loader_ext)
+static FASTAloader new_FASTAINDEX_loader(SEXP lkup, int load_descs,
+					 FASTAINDEX_loaderExt *loader_ext)
 {
 	FASTAloader loader;
 
@@ -133,9 +151,9 @@ static FASTAloader new_FASTAINFO_loader(SEXP lkup, int load_descs,
 		loader.lkup = INTEGER(lkup);
 		loader.lkup_length = LENGTH(lkup);
 	}
-	loader.load_desc_line = load_descs ? &FASTAINFO_load_desc_line : NULL;
-	loader.load_empty_seq = &FASTAINFO_load_empty_seq;
-	loader.load_seq_data = &FASTAINFO_load_seq_data;
+	loader.load_desc_line = load_descs ? &FASTAINDEX_load_desc_line : NULL;
+	loader.load_empty_seq = &FASTAINDEX_load_empty_seq;
+	loader.load_seq_data = &FASTAINDEX_load_seq_data;
 	loader.nrec = 0;
 	loader.ext = loader_ext;
 	return loader;
@@ -150,6 +168,14 @@ typedef struct fasta_loader_ext {
 	Chars_holder ans_elt_holder;
 } FASTA_loaderExt;
 
+static FASTA_loaderExt new_FASTA_loaderExt(SEXP ans)
+{
+	FASTA_loaderExt loader_ext;
+
+	loader_ext.ans_holder = hold_XVectorList(ans);
+	return loader_ext;
+}
+
 static void FASTA_load_empty_seq(FASTAloader *loader)
 {
 	FASTA_loaderExt *loader_ext;
@@ -157,8 +183,9 @@ static void FASTA_load_empty_seq(FASTAloader *loader)
 
 	loader_ext = loader->ext;
 	ans_elt_holder = &(loader_ext->ans_elt_holder);
-	*ans_elt_holder = get_elt_from_XRawList_holder(&(loader_ext->ans_holder),
-						 loader->nrec);
+	*ans_elt_holder = get_elt_from_XRawList_holder(
+					&(loader_ext->ans_holder),
+					loader->nrec);
 	ans_elt_holder->length = 0;
 	return;
 }
@@ -171,22 +198,14 @@ static void FASTA_load_seq_data(FASTAloader *loader,
 
 	loader_ext = loader->ext;
 	ans_elt_holder = &(loader_ext->ans_elt_holder);
-	/* ans_elt_holder->seq is a const char * so we need to cast it to
+	/* ans_elt_holder->ptr is a const char * so we need to cast it to
 	   char * in order to write to it */
-	memcpy((char *) ans_elt_holder->seq + ans_elt_holder->length,
-	       seq_data->seq, seq_data->length * sizeof(char));
+	memcpy((char *) ans_elt_holder->ptr + ans_elt_holder->length,
+	       seq_data->ptr, seq_data->length * sizeof(char));
 	ans_elt_holder->length += seq_data->length;
 	return;
 }
 
-static FASTA_loaderExt new_FASTA_loaderExt(SEXP ans)
-{
-	FASTA_loaderExt loader_ext;
-
-	loader_ext.ans_holder = hold_XVectorList(ans);
-	return loader_ext;
-}
-
 static FASTAloader new_FASTA_loader(SEXP lkup, FASTA_loaderExt *loader_ext)
 {
 	FASTAloader loader;
@@ -211,12 +230,12 @@ static int translate(Chars_holder *seq_data, const int *lkup, int lkup_length)
 	char *dest;
 	int nbinvalid, i, j, key, val;
 
-	/* seq_data->seq is a const char * so we need to cast it to
+	/* seq_data->ptr is a const char * so we need to cast it to
 	   char * before we can write to it */
-	dest = (char *) seq_data->seq;
+	dest = (char *) seq_data->ptr;
 	nbinvalid = j = 0;
 	for (i = 0; i < seq_data->length; i++) {
-		key = (unsigned char) seq_data->seq[i];
+		key = (unsigned char) seq_data->ptr[i];
 		if (key >= lkup_length || (val = lkup[key]) == NA_INTEGER) {
 			nbinvalid++;
 			continue;
@@ -231,13 +250,16 @@ static int translate(Chars_holder *seq_data, const int *lkup, int lkup_length)
  * Ignore empty lines and lines starting with 'FASTA_comment_markup' like in
  * the original Pearson FASTA format.
  */
-static const char *parse_FASTA_file(SEXP efp, int *recno, int *ninvalid,
-		int nrec, int skip, int seek_first_rec, FASTAloader *loader)
+static const char *parse_FASTA_file(SEXP efp,
+		int nrec, int skip, int seek_first_rec,
+		FASTAloader *loader,
+		int *recno, long long int *offset, long long int *ninvalid)
 {
-	int lineno, EOL_in_buf, EOL_in_prev_buf, ret_code,
+	int lineno, EOL_in_buf, EOL_in_prev_buf, ret_code, nbyte_in,
 	    FASTA_desc_markup_length, load_rec, is_new_rec;
 	char buf[IOBUF_SIZE];
 	Chars_holder data;
+	long long int prev_offset;
 
 	FASTA_desc_markup_length = strlen(FASTA_desc_markup);
 	load_rec = -1;
@@ -252,6 +274,14 @@ static const char *parse_FASTA_file(SEXP efp, int *recno, int *ninvalid,
 				 "from line %d", lineno);
 			return errmsg_buf;
 		}
+		if (EOL_in_buf) {
+			nbyte_in = strlen(buf);
+			data.length = delete_trailing_LF_or_CRLF(buf, nbyte_in);
+		} else {
+			data.length = nbyte_in = IOBUF_SIZE - 1;
+		}
+		prev_offset = *offset;
+		*offset += nbyte_in;
 		if (seek_first_rec) {
 			if (EOL_in_prev_buf
 			 && has_prefix(buf, FASTA_desc_markup)) {
@@ -260,17 +290,12 @@ static const char *parse_FASTA_file(SEXP efp, int *recno, int *ninvalid,
 				continue;
 			}
 		}
-		if (EOL_in_buf) {
-			data.length = delete_trailing_LF_or_CRLF(buf, -1);
-		} else {
-			data.length = IOBUF_SIZE - 1;
-		}
-		data.seq = buf;
+		data.ptr = buf;
 		if (!EOL_in_prev_buf)
 			goto parse_seq_data;
 		if (data.length == 0)
 			continue; // we ignore empty lines
-		if (has_prefix(data.seq, FASTA_comment_markup)) {
+		if (has_prefix(data.ptr, FASTA_comment_markup)) {
 			if (!EOL_in_buf) {
 				snprintf(errmsg_buf, sizeof(errmsg_buf),
 					 "cannot read line %d, "
@@ -280,7 +305,7 @@ static const char *parse_FASTA_file(SEXP efp, int *recno, int *ninvalid,
 			continue; // we ignore comment lines
 		}
 		buf[data.length] = '\0';
-		is_new_rec = has_prefix(data.seq, FASTA_desc_markup);
+		is_new_rec = has_prefix(data.ptr, FASTA_desc_markup);
 		if (is_new_rec) {
 			if (!EOL_in_buf) {
 				snprintf(errmsg_buf, sizeof(errmsg_buf),
@@ -293,9 +318,11 @@ static const char *parse_FASTA_file(SEXP efp, int *recno, int *ninvalid,
 				return NULL;
 			load_rec = load_rec && loader != NULL;
 			if (load_rec && loader->load_desc_line != NULL) {
-				data.seq += FASTA_desc_markup_length;
+				data.ptr += FASTA_desc_markup_length;
 				data.length -= FASTA_desc_markup_length;
-				loader->load_desc_line(loader, &data);
+				loader->load_desc_line(loader,
+						       *recno, prev_offset,
+						       &data);
 			}
 			if (load_rec && loader->load_empty_seq != NULL)
 				loader->load_empty_seq(loader);
@@ -327,89 +354,173 @@ static const char *parse_FASTA_file(SEXP efp, int *recno, int *ninvalid,
 	return NULL;
 }
 
-/* --- .Call ENTRY POINT --- */
-SEXP fasta_info(SEXP efp_list, SEXP nrec, SEXP skip, SEXP seek_first_rec,
-		SEXP use_names, SEXP lkup)
+static SEXP make_fasta_index_data_frame(const IntAE *recno_buf,
+					const IntAE *fileno_buf,
+					const LongLongIntAE *offset_buf,
+					const CharAEAE *desc_buf,
+					const IntAE *seqlength_buf)
 {
-	int nrec0, skip0, seek_rec0, load_descs, i, recno, ninvalid;
-	FASTAINFO_loaderExt loader_ext;
+	SEXP df, colnames, tmp;
+	int i;
+
+	PROTECT(df = NEW_LIST(5));
+
+	PROTECT(colnames = NEW_CHARACTER(5));
+	PROTECT(tmp = mkChar("recno"));
+	SET_STRING_ELT(colnames, 0, tmp);
+	UNPROTECT(1);
+	PROTECT(tmp = mkChar("fileno"));
+	SET_STRING_ELT(colnames, 1, tmp);
+	UNPROTECT(1);
+	PROTECT(tmp = mkChar("offset"));
+	SET_STRING_ELT(colnames, 2, tmp);
+	UNPROTECT(1);
+	PROTECT(tmp = mkChar("desc"));
+	SET_STRING_ELT(colnames, 3, tmp);
+	UNPROTECT(1);
+	PROTECT(tmp = mkChar("seqlength"));
+	SET_STRING_ELT(colnames, 4, tmp);
+	UNPROTECT(1);
+	SET_NAMES(df, colnames);
+	UNPROTECT(1);
+
+	PROTECT(tmp = new_INTEGER_from_IntAE(recno_buf));
+	SET_ELEMENT(df, 0, tmp);
+	UNPROTECT(1);
+
+	PROTECT(tmp = new_INTEGER_from_IntAE(fileno_buf));
+	SET_ELEMENT(df, 1, tmp);
+	UNPROTECT(1);
+
+	PROTECT(tmp = NEW_NUMERIC(LongLongIntAE_get_nelt(offset_buf)));
+	for (i = 0; i < LENGTH(tmp); i++)
+		REAL(tmp)[i] = (double) offset_buf->elts[i];
+	SET_ELEMENT(df, 2, tmp);
+	UNPROTECT(1);
+
+	PROTECT(tmp = new_CHARACTER_from_CharAEAE(desc_buf));
+	SET_ELEMENT(df, 3, tmp);
+	UNPROTECT(1);
+
+	PROTECT(tmp = new_INTEGER_from_IntAE(seqlength_buf));
+	SET_ELEMENT(df, 4, tmp);
+	UNPROTECT(1);
+
+	/* list_as_data_frame() performs IN-PLACE coercion */
+	list_as_data_frame(df, IntAE_get_nelt(recno_buf));
+	UNPROTECT(1);
+	return df;
+}
+
+/* --- .Call ENTRY POINT --- */
+SEXP fasta_index(SEXP efp_list,
+		 SEXP nrec, SEXP skip, SEXP seek_first_rec, SEXP lkup)
+{
+	int nrec0, skip0, seek_rec0, i, recno, old_nrec, new_nrec, k;
+	FASTAINDEX_loaderExt loader_ext;
 	FASTAloader loader;
-	SEXP efp, ans, ans_names;
+	IntAE *seqlength_buf, fileno_buf;
+	SEXP efp;
+	long long int offset, ninvalid;
 	const char *errmsg;
 
 	nrec0 = INTEGER(nrec)[0];
 	skip0 = INTEGER(skip)[0];
 	seek_rec0 = LOGICAL(seek_first_rec)[0];
-	load_descs = LOGICAL(use_names)[0];
-	loader_ext = new_FASTAINFO_loaderExt();
-	loader = new_FASTAINFO_loader(lkup, load_descs, &loader_ext);
-	recno = 0;
-	for (i = 0; i < LENGTH(efp_list); i++) {
+	loader_ext = new_FASTAINDEX_loaderExt();
+	loader = new_FASTAINDEX_loader(lkup, 1, &loader_ext);
+	seqlength_buf = &(loader_ext.seqlength_buf);
+	fileno_buf = new_IntAE(0, 0, 0);
+	for (i = recno = 0; i < LENGTH(efp_list); i++) {
 		efp = VECTOR_ELT(efp_list, i);
-		ninvalid = 0;
-		errmsg = parse_FASTA_file(efp, &recno, &ninvalid,
-					  nrec0, skip0, seek_rec0, &loader);
+		offset = ninvalid = 0LL;
+		errmsg = parse_FASTA_file(efp, nrec0, skip0, seek_rec0,
+					  &loader, &recno, &offset, &ninvalid);
 		if (errmsg != NULL)
 			error("reading FASTA file %s: %s",
 			      CHAR(STRING_ELT(GET_NAMES(efp_list), i)),
 			      errmsg_buf);
-		if (ninvalid != 0)
-			warning("reading FASTA file %s: ignored %d "
+		if (ninvalid != 0LL)
+			warning("reading FASTA file %s: ignored %lld "
 				"invalid one-letter sequence codes",
 				CHAR(STRING_ELT(GET_NAMES(efp_list), i)),
 				ninvalid);
+		old_nrec = IntAE_get_nelt(&fileno_buf);
+		new_nrec = IntAE_get_nelt(seqlength_buf);
+		for (k = old_nrec; k < new_nrec; k++)
+			IntAE_insert_at(&fileno_buf, k, i + 1);
 	}
-	PROTECT(ans = new_INTEGER_from_IntAE(&(loader_ext.seqlengths_buf)));
-	if (load_descs) {
-		PROTECT(ans_names =
-		    new_CHARACTER_from_CharAEAE(&(loader_ext.ans_names_buf)));
-		SET_NAMES(ans, ans_names);
-		UNPROTECT(1);
-	}
-	UNPROTECT(1);
-	return ans;
+	return make_fasta_index_data_frame(&(loader_ext.recno_buf),
+					   &fileno_buf,
+					   &(loader_ext.offset_buf),
+					   &(loader_ext.desc_buf),
+					   seqlength_buf);
 }
 
-/* --- .Call ENTRY POINT --- */
-SEXP read_fasta_in_XStringSet(SEXP efp_list, SEXP nrec, SEXP skip,
-		SEXP seek_first_rec,
-		SEXP use_names, SEXP elementType, SEXP lkup)
+/* --- .Call ENTRY POINT ---
+ * "FASTA blocks" are groups of consecutive FASTA records.
+ * Args:
+ *   seqlength:   Integer vector with 1 element per record to read. The
+ *                elements must be placed in the order that the records are
+ *                going to be read.
+ *   efp_list:    A list of N "External File Pointers" (see io_utils.c) with 1
+ *                element per input file. Files are going to be accessed from
+ *                first to last in the list.
+ *   nrec_list:   A list of N integer vectors (1 element per input file).
+ *                Each integer vector has 1 value per FASTA block, which is the
+ *                number of records in the block. IMPORTANT: Even if not
+ *                required, the blocks in each integer vector should preferably
+ *                be placed in the same order as in the file. This ensures that
+ *                the calls to ExternalFilePtr_seek() are always moving forward
+ *                in the file and are therefore efficient (moving backward on a
+ *                compressed file can be *extremely* slow).
+ *   offset_list: A list of N numeric vectors (1 element per input file) with
+ *                the same shape as 'nrec_list', i.e. each numeric vector has 1
+ *                value per FASTA block. This value is the offset of the block
+ *                (i.e. the offset of its first record) relative to the start
+ *                of the file. Measured in bytes.
+ *   elementType: The elementType of the XStringSet to return (its class is
+ *                inferred from this).
+ *   lkup:        Lookup table for encoding the incoming sequence bytes.
+ */
+SEXP read_XStringSet_from_fasta_blocks(SEXP seqlength,
+		SEXP efp_list, SEXP nrec_list, SEXP offset_list,
+		SEXP elementType, SEXP lkup)
 {
-	int nrec0, skip0, seek_rec0, i, recno, ninvalid;
-	SEXP efp, ans_width, ans_names, ans;
 	const char *element_type;
 	char classname[40];  /* longest string should be "DNAStringSet" */
+	SEXP ans, efp, nrec, offset;
 	FASTA_loaderExt loader_ext;
 	FASTAloader loader;
+	int i, j, nrec_j, recno;
+	long long int offset_j, ninvalid;
 
-	nrec0 = INTEGER(nrec)[0];
-	skip0 = INTEGER(skip)[0];
-	seek_rec0 = LOGICAL(seek_first_rec)[0];
-	PROTECT(ans_width = fasta_info(efp_list, nrec, skip, seek_first_rec,
-				       use_names, lkup));
-	PROTECT(ans_names = GET_NAMES(ans_width));
-	SET_NAMES(ans_width, R_NilValue);
 	element_type = CHAR(STRING_ELT(elementType, 0));
 	if (snprintf(classname, sizeof(classname), "%sSet", element_type)
 	    >= sizeof(classname))
 	{
-		UNPROTECT(2);
 		error("Biostrings internal error in "
-		      "read_fasta_in_XStringSet(): "
+		      "read_XStringSet_from_fasta_blocks(): "
 		      "'classname' buffer too small");
 	}
-	PROTECT(ans = alloc_XRawList(classname, element_type, ans_width));
-	_set_XStringSet_names(ans, ans_names);
+	PROTECT(ans = alloc_XRawList(classname, element_type, seqlength));
 	loader_ext = new_FASTA_loaderExt(ans);
 	loader = new_FASTA_loader(lkup, &loader_ext);
-	recno = ninvalid = 0;
 	for (i = 0; i < LENGTH(efp_list); i++) {
 		efp = VECTOR_ELT(efp_list, i);
-		ExternalFilePtr_rewind(efp);
-		parse_FASTA_file(efp, &recno, &ninvalid,
-				 nrec0, skip0, seek_rec0, &loader);
+		nrec = VECTOR_ELT(nrec_list, i);
+		offset = VECTOR_ELT(offset_list, i);
+		for (j = 0; j < LENGTH(nrec); j++) {
+			nrec_j = INTEGER(nrec)[j];
+			offset_j = llround(REAL(offset)[j]);
+			ExternalFilePtr_seek(efp, offset_j, SEEK_SET);
+			recno = 0;
+			ninvalid = 0LL;
+			parse_FASTA_file(efp, nrec_j, 0, 0,
+					 &loader, &recno, &offset_j, &ninvalid);
+		}
 	}
-	UNPROTECT(3);
+	UNPROTECT(1);
 	return ans;
 }
 
@@ -461,7 +572,7 @@ SEXP write_XStringSet_to_fasta(SEXP x, SEXP efp_list, SEXP width, SEXP lkup)
 			j2--;
 			Ocopy_bytes_from_i1i2_with_lkup(j1, j2,
 				buf, dest_nbytes,
-				X_elt.seq, X_elt.length,
+				X_elt.ptr, X_elt.length,
 				lkup0, lkup_length);
 			buf[dest_nbytes] = 0;
 			ExternalFilePtr_puts(efp, buf);
@@ -506,6 +617,14 @@ typedef struct fastqgeom_loader_ext {
 	int width;
 } FASTQGEOM_loaderExt;
 
+static FASTQGEOM_loaderExt new_FASTQGEOM_loaderExt()
+{
+	FASTQGEOM_loaderExt loader_ext;
+
+	loader_ext.width = NA_INTEGER;
+	return loader_ext;
+}
+
 static void FASTQGEOM_load_seq(FASTQloader *loader, const Chars_holder *seq)
 {
 	FASTQGEOM_loaderExt *loader_ext;
@@ -522,14 +641,6 @@ static void FASTQGEOM_load_seq(FASTQloader *loader, const Chars_holder *seq)
 	return;
 }
 
-static FASTQGEOM_loaderExt new_FASTQGEOM_loaderExt()
-{
-	FASTQGEOM_loaderExt loader_ext;
-
-	loader_ext.width = NA_INTEGER;
-	return loader_ext;
-}
-
 static FASTQloader new_FASTQGEOM_loader(FASTQGEOM_loaderExt *loader_ext)
 {
 	FASTQloader loader;
@@ -554,35 +665,6 @@ typedef struct fastq_loader_ext {
 	int lkup_length;
 } FASTQ_loaderExt;
 
-static void FASTQ_load_seqid(FASTQloader *loader, const Chars_holder *seqid)
-{
-	FASTQ_loaderExt *loader_ext;
-	CharAEAE *ans_names_buf;
-
-	loader_ext = loader->ext;
-	ans_names_buf = &(loader_ext->ans_names_buf);
-	// This works only because seqid->seq is nul-terminated!
-	append_string_to_CharAEAE(ans_names_buf, seqid->seq);
-	return;
-}
-
-static void FASTQ_load_seq(FASTQloader *loader, const Chars_holder *seq)
-{
-	FASTQ_loaderExt *loader_ext;
-	Chars_holder ans_elt_holder;
-
-	loader_ext = loader->ext;
-	ans_elt_holder = get_elt_from_XRawList_holder(&(loader_ext->ans_holder),
-						loader->nrec);
-	/* ans_elt_holder.seq is a const char * so we need to cast it to
-	   char * before we can write to it */
-	Ocopy_bytes_to_i1i2_with_lkup(0, ans_elt_holder.length - 1,
-		(char *) ans_elt_holder.seq, ans_elt_holder.length,
-		seq->seq, seq->length,
-		loader_ext->lkup, loader_ext->lkup_length);
-	return;
-}
-
 static FASTQ_loaderExt new_FASTQ_loaderExt(SEXP ans, SEXP lkup)
 {
 	FASTQ_loaderExt loader_ext;
@@ -600,6 +682,35 @@ static FASTQ_loaderExt new_FASTQ_loaderExt(SEXP ans, SEXP lkup)
 	return loader_ext;
 }
 
+static void FASTQ_load_seqid(FASTQloader *loader, const Chars_holder *seqid)
+{
+	FASTQ_loaderExt *loader_ext;
+	CharAEAE *ans_names_buf;
+
+	loader_ext = loader->ext;
+	ans_names_buf = &(loader_ext->ans_names_buf);
+	// This works only because seqid->ptr is nul-terminated!
+	append_string_to_CharAEAE(ans_names_buf, seqid->ptr);
+	return;
+}
+
+static void FASTQ_load_seq(FASTQloader *loader, const Chars_holder *seq)
+{
+	FASTQ_loaderExt *loader_ext;
+	Chars_holder ans_elt_holder;
+
+	loader_ext = loader->ext;
+	ans_elt_holder = get_elt_from_XRawList_holder(&(loader_ext->ans_holder),
+						loader->nrec);
+	/* ans_elt_holder.ptr is a const char * so we need to cast it to
+	   char * before we can write to it */
+	Ocopy_bytes_to_i1i2_with_lkup(0, ans_elt_holder.length - 1,
+		(char *) ans_elt_holder.ptr, ans_elt_holder.length,
+		seq->ptr, seq->length,
+		loader_ext->lkup, loader_ext->lkup_length);
+	return;
+}
+
 static FASTQloader new_FASTQ_loader(int load_seqids,
 				    FASTQ_loaderExt *loader_ext)
 {
@@ -614,12 +725,12 @@ static FASTQloader new_FASTQ_loader(int load_seqids,
 	return loader;
 }
 
-
 /*
  * Ignore empty lines.
  */
-static const char *parse_FASTQ_file(SEXP efp, int *recno,
-		int nrec, int skip, int seek_first_rec, FASTQloader *loader)
+static const char *parse_FASTQ_file(SEXP efp,
+		int nrec, int skip, int seek_first_rec,
+		FASTQloader *loader, int *recno)
 {
 	int lineno, EOL_in_buf, EOL_in_prev_buf, ret_code,
 	    FASTQ_line1_markup_length, FASTQ_line3_markup_length,
@@ -659,7 +770,7 @@ static const char *parse_FASTQ_file(SEXP efp, int *recno,
 		if (data.length == 0)
 			continue; // we ignore empty lines
 		buf[data.length] = '\0';
-		data.seq = buf;
+		data.ptr = buf;
 		lineinrecno++;
 		if (lineinrecno > 4)
 			lineinrecno = 1;
@@ -678,7 +789,7 @@ static const char *parse_FASTQ_file(SEXP efp, int *recno,
 			if (load_rec && nrec >= 0)
 				load_rec = *recno < skip + nrec;
 			if (load_rec && loader->load_seqid != NULL) {
-				data.seq += FASTQ_line1_markup_length;
+				data.ptr += FASTQ_line1_markup_length;
 				data.length -= FASTQ_line1_markup_length;
 				loader->load_seqid(loader, &data);
 			}
@@ -695,7 +806,7 @@ static const char *parse_FASTQ_file(SEXP efp, int *recno,
 				return errmsg_buf;
 			}
 			if (load_rec && loader->load_qualid != NULL) {
-				data.seq += FASTQ_line3_markup_length;
+				data.ptr += FASTQ_line3_markup_length;
 				data.length -= FASTQ_line3_markup_length;
 				loader->load_qualid(loader, &data);
 			}
@@ -734,8 +845,8 @@ SEXP fastq_geometry(SEXP efp_list, SEXP nrec, SEXP skip, SEXP seek_first_rec)
 	recno = 0;
 	for (i = 0; i < LENGTH(efp_list); i++) {
 		efp = VECTOR_ELT(efp_list, i);
-		errmsg = parse_FASTQ_file(efp, &recno,
-					  nrec0, skip0, seek_rec0, &loader);
+		errmsg = parse_FASTQ_file(efp, nrec0, skip0, seek_rec0,
+					  &loader, &recno);
 		if (errmsg != NULL)
 			error("reading FASTQ file %s: %s",
 			      CHAR(STRING_ELT(GET_NAMES(efp_list), i)),
@@ -749,7 +860,7 @@ SEXP fastq_geometry(SEXP efp_list, SEXP nrec, SEXP skip, SEXP seek_first_rec)
 }
 
 /* --- .Call ENTRY POINT --- */
-SEXP read_fastq_in_XStringSet(SEXP efp_list, SEXP nrec, SEXP skip,
+SEXP read_XStringSet_from_fastq(SEXP efp_list, SEXP nrec, SEXP skip,
 		SEXP seek_first_rec,
 		SEXP use_names, SEXP elementType, SEXP lkup)
 {
@@ -771,7 +882,7 @@ SEXP read_fastq_in_XStringSet(SEXP efp_list, SEXP nrec, SEXP skip,
 	if (ans_length != 0) {
 		if (INTEGER(ans_geom)[1] == NA_INTEGER) {
 			UNPROTECT(2);
-			error("read_fastq_in_XStringSet(): FASTQ files with "
+			error("read_XStringSet_from_fastq(): FASTQ files with "
 			      "variable sequence lengths are not supported yet");
 		}
 		for (recno = 0; recno < ans_length; recno++)
@@ -783,7 +894,7 @@ SEXP read_fastq_in_XStringSet(SEXP efp_list, SEXP nrec, SEXP skip,
 	{
 		UNPROTECT(2);
 		error("Biostrings internal error in "
-		      "read_fastq_in_XStringSet(): "
+		      "read_XStringSet_from_fastq(): "
 		      "'classname' buffer too small");
 	}
 	PROTECT(ans = alloc_XRawList(classname, element_type, ans_width));
@@ -793,8 +904,8 @@ SEXP read_fastq_in_XStringSet(SEXP efp_list, SEXP nrec, SEXP skip,
 	for (i = 0; i < LENGTH(efp_list); i++) {
 		efp = VECTOR_ELT(efp_list, i);
 		ExternalFilePtr_rewind(efp);
-		parse_FASTQ_file(efp, &recno,
-				 nrec0, skip0, seek_rec0, &loader);
+		parse_FASTQ_file(efp, nrec0, skip0, seek_rec0,
+				 &loader, &recno);
 	}
 	if (load_seqids) {
 		PROTECT(ans_names =
@@ -865,7 +976,7 @@ static void write_FASTQ_qual(SEXP efp, int seqlen,
 	if (Q_elt.length != seqlen)
 		error("'x' and 'quality' must have the same width");
 	for (j = 0; j < seqlen; j++)
-		ExternalFilePtr_putc(efp, (int) *(Q_elt.seq++));
+		ExternalFilePtr_putc(efp, (int) *(Q_elt.ptr++));
 	ExternalFilePtr_puts(efp, "\n");
 }
 
@@ -913,7 +1024,7 @@ SEXP write_XStringSet_to_fastq(SEXP x, SEXP efp_list, SEXP qualities, SEXP lkup)
 		X_elt = _get_elt_from_XStringSet_holder(&X, i);
 		Ocopy_bytes_from_i1i2_with_lkup(0, X_elt.length - 1,
 			buf, X_elt.length,
-			X_elt.seq, X_elt.length,
+			X_elt.ptr, X_elt.length,
 			lkup0, lkup_length);
 		buf[X_elt.length] = 0;
 		write_FASTQ_id(efp, FASTQ_line1_markup, id);
