@@ -5,8 +5,9 @@
  *                                                                          *
  ****************************************************************************/
 #include "Biostrings.h"
-#include "IRanges_interface.h"
 #include "XVector_interface.h"
+#include "IRanges_interface.h"
+#include "S4Vectors_interface.h"
 
 static int debug = 0;
 
@@ -274,7 +275,7 @@ SEXP match_PDict3Parts_XStringViews(SEXP pptb, SEXP pdict_head, SEXP pdict_tail,
 		view_offset = *view_start - 1;
 		if (view_offset < 0 || view_offset + *view_width > S.length)
 			error("'subject' has \"out of limits\" views");
-		S_view.seq = S.seq + view_offset;
+		S_view.ptr = S.ptr + view_offset;
 		S_view.length = *view_width;
 		match_pdict(pptb, &headtail, &S_view,
 			    max_mismatch, min_mismatch, fixed,
@@ -402,7 +403,7 @@ static SEXP vwhich_XStringSet_XStringSet(SEXP pattern,
 			_drop_reported_matches();
 		}
 	}
-	return new_LIST_from_IntAEAE(&ans_buf, 1);
+	return new_LIST_from_IntAEAE(&ans_buf, 0);
 }
 
 static SEXP vcount_PDict3Parts_XStringSet(SEXP pptb, HeadTail *headtail,
