@@ -5,7 +5,7 @@
  *                                                                          *
  ****************************************************************************/
 #include "Biostrings.h"
-#include "IRanges_interface.h"
+#include "S4Vectors_interface.h"
 #include <S.h> /* for Salloc() */
 
 #include <limits.h> /* for ULONG_MAX */
@@ -471,7 +471,7 @@ static void preprocess_H(const Chars_holder *H,
 	const char *c;
 	BitMatrix *bmbuf;
 
-	for (j = 0, c = H->seq + H->length - 1; j < H->length; j++, c--) {
+	for (j = 0, c = H->ptr + H->length - 1; j < H->length; j++, c--) {
 		offset = byte2offset->byte2code[(unsigned char) *c];
 		if (offset == NA_INTEGER)
 			error("preprocess_H(): don't know how to handle "
@@ -495,7 +495,7 @@ static void preprocess_T(const Chars_holder *T,
 	const char *c;
 	BitMatrix *bmbuf;
 
-	for (j = 0, c = T->seq; j < T->length; j++, c++) {
+	for (j = 0, c = T->ptr; j < T->length; j++, c++) {
 		offset = byte2offset->byte2code[(unsigned char) *c];
 		if (offset == NA_INTEGER)
 			error("preprocess_T(): don't know how to handle "
@@ -559,7 +559,7 @@ static BitCol match_ppheadtail_for_loc(HeadTail *headtail, int tb_width,
 		// 'j2' should be a safe location in 'S' because we call
 		// match_ppheadtail_for_loc() only when 'tb_end' is guaranteed
 		// not to be too close to 'S' boundaries.
-		s = S->seq[j2];
+		s = S->ptr[j2];
 		offset = headtail->ppheadtail.byte2offset.byte2code[(unsigned char) s];
 		if (offset == NA_INTEGER) {
 			_BitMatrix_Rrot1(nmis_bmbuf);
@@ -577,7 +577,7 @@ static BitCol match_ppheadtail_for_loc(HeadTail *headtail, int tb_width,
 		// 'j2' should be a safe location in 'S' because we call
 		// match_ppheadtail_for_loc() only when 'tb_end' is guaranteed
 		// not to be too close from 'S' boundaries.
-		s = S->seq[j2];
+		s = S->ptr[j2];
 		offset = headtail->ppheadtail.byte2offset.byte2code[(unsigned char) s];
 		if (offset == NA_INTEGER) {
 			_BitMatrix_Rrot1(nmis_bmbuf);
@@ -881,7 +881,7 @@ dict0 <- DNAStringSet(drosophila2probe$sequence, end=12)
 pdict6 <- PDict(dict0, tb.start=4, tb.end=9)
 
 mi6 <- matchPDict(pdict6, chr3R_50000, max.mismatch=1)
-sum(countIndex(mi6))  # 17896
+sum(elementLengths(mi6))  # 17896
 
 		if (headtail->ppheadtail.is_init
 		 && IntAE_get_nelt(tb_end_buf) >= 40) {
